from typing import Any, Dict


class NotturnoASGI:
    def __init__(self, router):
        self.router = router

    async def _http(self, scope: Dict[str, Any], receive: Any, send: Any):
        pass

    async def _websocket(self, scope: Dict[str, Any], receive: Any, send: Any):
        pass

    async def _lifespan(self, scope: Dict[str, Any], receive: Any, send: Any):
        pass



import ssl
import traceback

import anyio
from anyio.streams.tls import TLSListener, TLSStream
from anyio._backends._asyncio import SocketStream as AIOSocketStream
from anyio._backends._trio import SocketStream as TrioSocketStream

class NoctServ:
    def __init__(self, handler):
        self.handler = handler
        self.server_hide = None
        self.connections = {}

    def parse_http_message(self, http_message):
        if not http_message.strip():
            return None, {}, ""

        lines = http_message.splitlines()
        start_line = lines[0] if lines else ""
        parts = start_line.split()

        if len(parts) < 3:
            return None, {}, ""

        method = parts[0]
        path = parts[1]
        http_version = parts[2]
        headers = {}
        header_lines = []

        for line in lines[1:]:
            if line == "":
                break
            header_lines.append(line)

        for header in header_lines:
            key, value = header.split(": ", 1)
            headers[key] = value

        body_start_index = len(header_lines) + 2
        body = "\n".join(lines[body_start_index:])

        return method, path, headers, body, http_version

    async def __handle(
        self,
        client: TLSStream
        | AIOSocketStream
        | TrioSocketStream,
    ):
        conn_type = None
        try:
            
            async with client:
                data = await client.receive(1024)
                reqline = data.decode()
                method, path, headers, body, http_version = self.parse_http_message(
                    reqline
                )
                if not headers.get("Upgrade") == "websocket":
                    conn_type = "http"
                    response = await self.handler._native_http_handle(
                        method, path, headers, body, http_version
                    )
                    await client.send(response.encode())
                else:
                    conn_type = "websocket"
                    await self.handler._native_ws_handle(
                        client, path, headers, body, http_version
                    )
        except (ssl.SSLError, anyio.BrokenResourceError, Exception) as e:
            if isinstance(e, ssl.SSLError):
                if e.reason == "APPLICATION_DATA_AFTER_CLOSE_NOTIFY":
                    print("SSL error: application data after close notify")
                elif e.reason == "UNEXPECTED_EOF_WHILE_READING":
                    print("SSL error: unexpected EOF while reading")
            elif not isinstance(e, anyio.BrokenResourceError) and isinstance(
                e, Exception
            ):
                print(traceback.format_exc())
                if conn_type == "http":
                    await self.send_error_response(client, 500, "Internal Server Error")

    async def send_error_response(self, client, status_code, message):
        response = (
            f"HTTP/1.1 {status_code} {message}\r\n"
            "Content-Type: text/plain\r\n"
            f"Content-Length: {len(message)}\r\n"
            "\r\n"
            f"{message}"
        )
        await client.send(response.encode())

    async def serve(
        self,
        host: str = "127.0.0.1",
        port: int = 8765,
        server_hide: bool = False,
        use_ssl: bool = False,
        certfile: str = "cert.pem",
        keyfile: str = "key.pem",
    ):
        self.server_hide = server_hide
        listener = await anyio.create_tcp_listener(local_host=host, local_port=port)
        self.ssl = use_ssl
        if use_ssl:
            context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
            context.load_cert_chain(certfile=certfile, keyfile=keyfile)
            context.set_alpn_protocols(["http/1.1", "h2", "h3"])
            listener = TLSListener(listener, context, standard_compatible=False)
            print(f"Server is running on https://{host}:{port}")
        else:
            print(f"Server is running on http://{host}:{port}")
        await listener.serve(self.__handle)



from urllib.parse import parse_qsl

try:
    from fast_query_parsers import parse_query_string
except ModuleNotFoundError:
    parse_query_string = None

def parse_qs(qs: str) -> dict[str, str]:
    if parse_query_string:
        return dict(parse_query_string(qs, "&"))
    return dict(parse_qsl(qs))


import re2

class RegExpRouter:
    def __init__(self, root_path: str = ""):
        self.routes = {}
        self.root_path = root_path
        self.compiled_regex = {}

    def add_route(self, method, pattern, handler):
        pattern = self.root_path + pattern
        if pattern == "/" or pattern == "":
            regex_pattern = r"^/$"
        else:
            regex_pattern = re2.sub(
                r":(\w+)", r"(?P<\1>[^/]+)", pattern
            )
            regex_pattern = re2.sub(r"\*", r"(.+)", regex_pattern)

        if method not in self.routes:
            self.routes[method] = []

        self.routes[method].append((regex_pattern, handler))
        self.compile_routes(method)

    def compile_routes(self, method):
        combined_pattern = "|".join(
            f"(?P<route_{index}_{method}>{pattern})"
            for index, (pattern, _) in enumerate(self.routes[method])
        )
        self.compiled_regex[method] = re2.compile(combined_pattern)

    def match(self, path):
        result = {}
        for method, patterns in self.routes.items():
            if method not in self.compiled_regex:
                continue

            match = self.compiled_regex[method].match(path)
            if match:
                for index, (pattern, handler) in enumerate(patterns):
                    if match.group(f"route_{index}_{method}"):
                        params = {
                            key: match.group(key)
                            for key in match.groupdict()
                            if key
                            not in [f"route_{i}_{method}" for i in range(len(patterns))]
                        }
                        if not params:
                            params = {}
                        result[method] = {
                            "func": handler,
                            "params": params,
                        }
            else:
                result[method] = None
        return result if result else None

    def combine(self, other_router):
        for method, routes in other_router.routes.items():
            for pattern, handler in routes:
                self.add_route(method, pattern, handler)



from typing import Any, Dict, Optional

from pydantic import BaseModel, Field

from ..utils.url import URL


class Request(BaseModel):
    method: str = Field(description="HTTP Method")
    url: URL = Field(description="Request URL")
    headers: Optional[Dict[str, str]] = Field(
        default_factory=dict, description="Request Header"
    )
    query: Optional[Dict[str, Any]] = Field(
        default_factory=dict, description="Query Params"
    )
    body: Optional[Any] = Field(default=None, description="Request Body")



from .app import Notturno
from .core.router.regexp import RegExpRouter


class Gear(Notturno):
    def __init__(self, root_path: str = "", async_backend="asyncio"):
        super().__init__(async_backend)
        self._router = RegExpRouter(root_path=root_path)
        self._internal_router = RegExpRouter(root_path=root_path)



from typing import Annotated, Any

from yarl import URL as YarlURL
from pydantic import BaseModel, PlainValidator, ValidationInfo

def validate(v: Any, info: ValidationInfo) -> YarlURL:
    if isinstance(v, YarlURL):
        ans = v
    elif isinstance(v, str):
        ans = YarlURL(v)
    else:
        raise TypeError(
            f"Expected yarl.URL, got {type(v)}"
        )
    return ans

URL = Annotated[
    YarlURL,
    PlainValidator(validate)
]


try:
    import orjson
except ModuleNotFoundError:
    orjson = None
    import json

def dumps(obj: dict) -> bytes:
    if orjson:
        return orjson.dumps(obj)
    return json.dumps(obj).encode()

def loads(obj: str | bytes | bytearray) -> bytes:
    if orjson:
        return orjson.loads(obj)
    return json.loads(obj)


from pydantic import BaseModel, Field
from typing import Optional, Dict, Any, Union

class Response(BaseModel):
    body: Optional[Union[Dict[str, Any], Any, str, bytes]] = Field(default="", description="Response Body")
    headers: Optional[Dict[str, str]] = Field(default_factory=dict, description="Response Header")
    status_code: Optional[int] = Field(default=200, description="HTTP Status Code")
    content_type: Optional[Union[str, None]] = Field(default=None)



import asyncio
import inspect
from functools import partial, wraps
from http.client import responses
from typing import Any, Callable, Dict

import anyio
from yarl import URL

from .core.http.serv import NoctServ
from .core.router.regexp import RegExpRouter
from .models.request import Request
from .models.response import Response
from .models.websocket import WebSocket
from .utils import jsonenc
from .utils.query import parse_qs


class Notturno:
    def __init__(self, async_backend: str = "asyncio"):
        self._router = RegExpRouter()
        self.dependencies = {}
        self._internal_router = RegExpRouter()
        self.http = NoctServ(self)
        self.async_backend = async_backend
        self.__server_hide = None
        self.__is_main = self.__is_non_gear()

    def __is_non_gear(self):
        if isinstance(self, Notturno) and type(self) is Notturno:
            return True
        else:
            return False

    def merge_route(self, cls):
        if isinstance(cls, Notturno):
            self._router.combine(cls._router)
            self._internal_router.combine(cls._internal_router)
        else:
            raise TypeError(
                f"Notturno.Notturno or Notturno.Gear required, but got {cls.__class__}"
            )

    def add_dependency(self, name: str, instance):
        self.dependencies[name] = instance

    def di(self, *dependency_names):
        def decorator(func):
            @wraps(func)
            async def wrapper(*args, **kwargs):
                for name in dependency_names:
                    if name in self.dependencies:
                        kwargs[name] = self.dependencies[name]
                return await func(*args, **kwargs)

            return wrapper

        return decorator

    async def __asgi_http_handle(self, scope: Dict[str, Any], receive: Any, send: Any):
        route, params = await self.resolve(scope["method"], scope["path"])
        if not route:
            await send(
                {
                    "type": "http.response.start",
                    "status": 404,
                    "headers": [[b"Content-Type", b"text/plain"]],
                }
            )
            await send(
                {
                    "type": "http.response.body",
                    "body": b"Not Found",
                    "more_body": False,
                }
            )
            return
        response_body = b""
        while True:
            message = await receive()
            response_body += message.get("body", b"")
            if not message.get("more_body", False):
                break
        response_body_string = response_body.decode("utf-8")
        headers = {
            key.decode("utf-8"): value.decode("utf-8")
            for key, value in scope["headers"]
        }
        arg_name = await self.__route(func=route, is_type=Request)
        if arg_name:
            req = Request(
                method=scope["method"],
                url=f"{scope['scheme']}://{headers['host']}{scope['path']}",
                headers=headers,
                query=parse_qs(scope["query_string"]),
                body=response_body_string,
            )
            params[arg_name] = req
        if asyncio.iscoroutinefunction(route):
            resp = await route(**params)
        else:
            resp = route(**params)
        if isinstance(resp, Response):
            content_type = None
            if isinstance(resp.body, dict):
                resp.body = jsonenc.dumps(resp)
                content_type = "application/json"
            elif isinstance(resp.body, list):
                resp.body = b"".join([s.encode("utf-8") for s in resp.body])
                content_type = "application/json"
            elif isinstance(resp.body, str):
                resp.body = resp.body.encode()
                content_type = "text/plain"
            elif isinstance(resp.body, bytes):
                content_type = "application/octet-stream"
            elif isinstance(resp.body, int) or isinstance(resp.body, float):
                resp.body = resp.body.to_bytes(4, byteorder="big")
                content_type = "text/plain"
            else:
                content_type = "application/octet-stream"
            if not resp.headers.get("Content-Type"):
                if resp.content_type:
                    resp.headers["Content-Type"] = resp.content_type
                elif content_type:
                    resp.headers["Content-Type"] = content_type
            await send(
                {
                    "type": "http.response.start",
                    "status": resp.status_code,
                    "headers": [
                        [key.encode("utf-8"), value.encode("utf-8")]
                        for key, value in resp.headers.items()
                    ],
                }
            )
            await send(
                {
                    "type": "http.response.body",
                    "body": resp.body,
                    "more_body": False,
                }
            )
        elif isinstance(resp, dict):
            await send(
                {
                    "type": "http.response.start",
                    "status": 200,
                    "headers": [[b"content-type", b"application/json"]],
                }
            )
            await send(
                {
                    "type": "http.response.body",
                    "body": jsonenc.dumps(resp),
                    "more_body": False,
                }
            )

        elif isinstance(resp, str):
            await send(
                {
                    "type": "http.response.start",
                    "status": 200,
                    "headers": [[b"content-type", b"text/plain"]],
                }
            )
            await send(
                {
                    "type": "http.response.body",
                    "body": resp.encode(),
                    "more_body": False,
                }
            )
        elif not resp:
            await send(
                {
                    "type": "http.response.start",
                    "status": 200,
                    "headers": [[b"content-type", b"text/plain"]],
                }
            )
            await send(
                {
                    "type": "http.response.body",
                    "body": b"",
                    "more_body": False,
                }
            )
        if isinstance(resp, tuple):
            if len(resp) == 2:
                if isinstance(resp[0], dict) and isinstance(resp[1], int):
                    await send(
                        {
                            "type": "http.response.start",
                            "status": resp[1],
                            "headers": [[b"content-type", b"application/json"]],
                        }
                    )
                    await send(
                        {
                            "type": "http.response.body",
                            "body": jsonenc.dumps(resp[0]),
                            "more_body": False,
                        }
                    )
                elif isinstance(resp[0], str) and isinstance(resp[1], int):
                    await send(
                        {
                            "type": "http.response.start",
                            "status": resp[1],
                            "headers": [[b"content-type", b"text/plain"]],
                        }
                    )
                    await send(
                        {
                            "type": "http.response.body",
                            "body": resp[0].encode(),
                            "more_body": False,
                        }
                    )

    async def __call__(self, scope: Dict[str, Any], receive: Any, send: Any):
        if not self.__is_main:
            raise TypeError(
                "You cannot start a server with anything Notturno.Gear as your main."
            )
        if scope["type"] == "http":
            await self.__asgi_http_handle(scope, receive, send)
        elif scope["type"] == "websocket":
            pass

    def __normalize_path(self, path: str) -> str:
        return path.rstrip("/")

    async def __route(self, func: Callable, is_type) -> Any:
        signature = inspect.signature(func)
        arg_name = None
        for param_name, param in signature.parameters.items():
            if param.annotation is is_type:
                arg_name = param_name
                break
        # if request_arg_name is not None:
        #    kwargs[request_arg_name] = request_value

        # if asyncio.iscoroutinefunction(func):
        #    return await func(*args, **kwargs)
        # else:
        #    return func(*args, **kwargs)
        return arg_name

    async def _native_http_handle(
        self, method: str, path, headers, body, http_version
    ) -> str:
        route, params = await self.resolve(method.upper(), path)
        if not route:
            msg = "Not Found"
            response = (
                f"HTTP/1.1 404 NotFound\r\n"
                "Content-Type: text/plain\r\n"
                f"Content-Length: {len(msg)}\r\n"
                "\r\n"
                f"{msg}"
            )
            return response
        url = URL(f"{'https' if self.ssl else 'http'}://{headers['Host']}/{path}")

        arg_name = await self.__route(func=route, is_type=Request)
        if arg_name:
            req = Request(
                method=method.upper(),
                url=url,
                headers=headers,
                query={key: url.query.getlist(key) for key in url.query.keys()},
                body=body,
            )
            params[arg_name] = req
        if asyncio.iscoroutinefunction(route):
            resp = await route(**params)
        else:
            resp = route(**params)
        if isinstance(resp, Response):
            content_type = None
            if isinstance(resp.body, dict):
                resp.body = jsonenc.dumps(resp)
                content_type = "application/json"
            elif isinstance(resp.body, list):
                content_type = "application/json"
            elif isinstance(resp.body, str):
                content_type = "text/plain"
            elif isinstance(resp.body, bytes):
                content_type = "application/octet-stream"
            elif isinstance(resp.body, int) or isinstance(resp.body, float):
                content_type = "text/plain"
            else:
                content_type = "application/octet-stream"
            resp_desc = responses.get(resp.status_code)
            if not resp.headers.get("Content-Length"):
                resp.headers["Content-Length"] = len(resp.body)
            if not resp.headers.get("Content-Type"):
                if resp.content_type:
                    resp.headers["Content-Type"] = resp.content_type
                elif content_type:
                    resp.headers["Content-Type"] = content_type
            if not self.__server_hide:
                resp.headers["Server"] = "NoctServ/0.1.0"
            else:
                resp.headers["Server"] = "NoctServ"
            headers = [f"{key}: {value}" for key, value in resp.headers.items()]
            response = (
                f"HTTP/1.1 {resp.status_code} {resp_desc if resp_desc else 'UNKNOWN'}\r\n"
                f"{'\r\n'.join(headers)}\r\n"
                "\r\n"
                f"{resp.body}"
            )
        elif isinstance(resp, dict):
            dumped = jsonenc.dumps(resp)
            response = (
                "HTTP/1.1 200 OK\r\n"
                f"Content-Length: {len(dumped)}\r\n"
                "Content-Type: application/json\r\n"
                "\r\n"
                f"{dumped}"
            )
        return response

    async def _native_ws_handle(self, client, path, headers, body, http_version):
        if "Sec-WebSocket-Key" not in headers:
            return "HTTP/1.1 400 Bad Request\r\nServer: NoctServe\r\n\r\nBad Request"
        route, params = await self.__resolve_internal("WS", path)
        if not route:
            msg = "Not Found"
            response = (
                f"HTTP/1.1 404 NotFound\r\n"
                "Content-Type: text/plain\r\n"
                f"Content-Length: {len(msg)}\r\n"
                "\r\n"
                f"{msg}"
            )
            await client.send(response.encode())
            await client.aclose()
            return
        raise NotImplementedError("Websocket Native Support is Non-Ready :(")
        ws = WebSocket(path, headers, http_version)
        ws._is_native = True
        ws._webkey = "Sec-WebSocket-Key"
        ws._client = client
        arg_name = await self.__route(func=route, is_type=WebSocket)
        params[arg_name] = ws
        if asyncio.iscoroutinefunction(route):
            await route(**params)
        else:
            raise TypeError("Websocket is Only to use in coroutine function.")

    async def resolve(
        self, method: str, path: str
    ) -> tuple[None, dict] | tuple[Any | None, dict]:
        route = self._router.match(path)
        if not route or not route.get(method):
            return (None, None)
        return (route.get(method)["func"], route.get(method)["params"])

    def route(self, route: str, method: list = ["GET"]):
        def decorator(func):
            route_normalized = self.__normalize_path(route)
            if isinstance(func, staticmethod):
                func = func.__func__
            func._router_method = method
            for m in method:
                met = m.upper()
                self._router.add_route(met, route_normalized, func)
            return func

        return decorator

    async def __resolve_internal(
        self, method: str, path: str
    ) -> tuple[None, dict] | tuple[Any | None, dict]:
        route = self._internal_router.match(path)
        if not route:
            return (None, None)
        return (route.get(method)["func"], route.get(method)["params"])

    def serve(
        self,
        host: str = "127.0.0.1",
        port: int = 8765,
        hide_server_version: bool = True,
        ssl: bool = False,
        certfile: str = "cert.pem",
        keyfile: str = "key.pem",
    ) -> None:
        if not self.__is_main:
            raise TypeError(
                "You cannot start a server with anything Notturno.Gear as your main."
            )
        self.__server_hide = hide_server_version
        self.ssl = ssl
        anyio.run(
            partial(
                self.http.serve,
                host=host,
                port=port,
                server_hide=hide_server_version,
                use_ssl=ssl,
                certfile=certfile,
                keyfile=keyfile,
            ),
            backend=self.async_backend,
        )

    def get(self, route: str):
        return self.route(route, method=["GET"])

    def post(self, route: str):
        return self.route(route, method=["POST"])

    def ws(self, route: str):
        def decorator(func):
            route_normalized = self.__normalize_path(route)
            if isinstance(func, staticmethod):
                func = func.__func__
            func._router_method = "WS"
            self._internal_router.add_route("WS", route_normalized, func)
            return func

        return decorator



import base64
import hashlib
import struct

from anyio.streams.tls import TLSStream
from anyio._backends._asyncio import SocketStream as AIOSocketStream
from anyio._backends._trio import SocketStream as TrioSocketStream

from ..exceptions import WebsocketClosed


class WebSocket:
    def __init__(self, path: str, headers: str, http_version: str):
        self._webkey = None
        self._client: TLSStream | AIOSocketStream | TrioSocketStream = None
        self._is_native = True

        self.path = path
        self.headers = headers
        self.http_version = http_version

        self.send = None
        self.receive = None

    async def accept(self):
        if self._is_native:
            webaccept = base64.b64encode(
                hashlib.sha1(
                    (self._webkey + "258EAFA5-E914-47DA-95CA-C5AB0DC85B11").encode()
                ).digest()
            ).decode()
            headers = [f"{key}: {value}" for key, value in {
                "Upgrade": "websocket",
                "Connection": "Upgrade",
                "Sec-WebSocket-Accept": f"{webaccept}",
                "Sec-WebSocket-Version": "13"
            }.items()]
            response = (
                "HTTP/1.1 101 Switching Protocols\r\n"
                f"{'\r\n'.join(headers)}\r\n"
            )
            await self._client.send(response.encode())
        else:
            await self.send({
                "type": "websocket.accept",
            })

    async def recv(self):
        if self._is_native:
            data = await self._client.receive(2)
            if len(data) < 2:
                raise WebsocketClosed
            
            length = data[1] & 127
            if length == 126:
                length_data = await self._client.receive(2)
                length = struct.unpack(">H", length_data)[0]
            elif length == 127:
                length_data = await self._client.receive(8)
                length = struct.unpack(">Q", length_data)[0]

            mask = await self._client.receive(4)

            message = await self._client.receive(length)

            decoded_message = bytearray(
                b ^ mask[i % 4] for i, b in enumerate(message)
            )

            try:
                return decoded_message.decode()
            except UnicodeDecodeError:
                return None
        else:
            message = await self.receive()
            if message["type"] == "websocket.disconnect":
                raise WebsocketClosed
            elif message["type"] == "websocket.receive":
                return message["text"]
            raise Exception("Invalid message type")
        
    async def send(self, message: str | bytes):
        if self._is_native:
            if isinstance(message, str):
                message = message.encode()
            length = len(message)
            if length <= 125:
                frame = struct.pack("B", 129) + struct.pack("B", length) + message
            elif length >= 126 and length <= 65535:
                frame = (
                    struct.pack("B", 129)
                    + struct.pack("B", 126)
                    + struct.pack(">H", length)
                    + message
                )
            else:
                raise Exception("Message too long")
            await self._client.send(frame)
        else:
            tmpl = {
                "type": "websocket.send",
            }
            if isinstance(message, bytes):
                tmpl["bytes"] = message
            else:
                tmpl["text"] = message
            await self.send(tmpl)
        
    async def close(self):
        if self._is_native:
            await self.client.aclose()
        else:
            await self.send({
                "type": "websocket.close",
            })